import 'package:polymer/polymer.dart';
import 'package:highcharts_options/chart_options.dart' as hc;
import 'package:polymer_highcharts/highcharts_series.dart';
import 'dart:html';
import 'dart:js';

@CustomTag('highcharts-polymer')
class HighchartsPolymerComponent extends PolymerElement {
  
  @published hc.HighChart chartOptions = new hc.HighChart();
  
  @published String chartTitle;
  @published String subtitle;
  @published bool creditsEnabled;
  @published String type;
  @published String borderColor;
  @published int borderWidth;
  @published int borderRadius;
  @published String backgroundColor;
  @published Map plotOptions;
  
  DivElement mainDiv;
  bool _domReady = false;
  bool _propertiesDirty = false;
  bool _seriesDirty = false;
  bool _chartCreated = false;
  
  bool _updateableSeries = false;
    
  Map<String, HighchartsSeries> _seriesDictionary = new Map<String, HighchartsSeries> ();
  List<HighchartsSeries> _previousSeries = new List<HighchartsSeries> ();
  
  List<Map> _pendingSeries = new List<Map> ();
  hc.XAxis _xAxis;
  hc.YAxis _yAxis;
  
  JsObject jsHighchart = null;
  
  factory HighchartsPolymerComponent() => new Element.tag('highcharts-polymer');
  HighchartsPolymerComponent.created () : super.created ();
  
  void titleChanged (String oldValue) => _invalidateChartOptions();
  void subtitleChanged (String oldValue) => _invalidateChartOptions();
  void creditsEnabledChanged (bool oldValue) => _invalidateChartOptions();
  void typeChanged (String oldValue) => _invalidateChartOptions();
  void borderColorChanged (String oldValue) => _invalidateChartOptions();
  void borderWidthChanged (int oldValue) => _invalidateChartOptions();
  void borderRadiusChanged (int oldValue) => _invalidateChartOptions();
  void plotOptionsChanged (int oldValue) => _invalidateChartOptions();
  
  bool _chartOptionsDirty = false;
  
  void _invalidateChartOptions () {
    _chartOptionsDirty = true;
    _invalidateProperties();
  }
  
  void _commitChartOptions () {
    if (chartOptions != null) {
      if (title != null) {
        chartOptions.title = (new hc.Title ()..text = chartTitle);
      } 
      if (subtitle != null) {
        chartOptions.subtitle = (new hc.Subtitle ()..text = subtitle);
      } 
      if (creditsEnabled != null) {
        chartOptions.credits = (new hc.Credits()..enabled = creditsEnabled);
      }
      if (chartOptions.chart == null) {
        chartOptions.chart = new hc.Chart();
      }
      if (type != null) {
        chartOptions.chart.type = type;
      }
      if (borderColor != null) {
        chartOptions.chart.borderColor = borderColor;
      }
      if (borderWidth != null) {
        chartOptions.chart.borderWidth = borderWidth;
      }
      if (borderRadius != null) {
        chartOptions.chart.borderRadius = borderRadius;
      }
      if (plotOptions != null) {
        if (chartOptions.plotOptions == null) {
          chartOptions.plotOptions = new hc.PlotOptions();
        }
        chartOptions.plotOptions.moreOptions = plotOptions;
      }
      _chartOptionsDirty = false;
    }
    
  }
  
  @override 
  void attached () {
    mainDiv = $['hcDiv'] as DivElement;
    _domReady = true;
    _chartOptionsDirty = true;
    _invalidateProperties ();
  }
  
  List<HighchartsSeries> _findSeriesInDOM () {
    List<HighchartsSeries> out = new List<HighchartsSeries> ();
    List<Element> elements = this.querySelectorAll('highcharts-series');
    elements.forEach((Element e) {
      out.add(e as HighchartsSeries);
    });
    return out;
  }
  
  void notifyNewSeries () {
    _seriesDirty = true;
    _invalidateProperties();
  }
  
  bool _updateable (List<HighchartsSeries> seriesInDOM, List<HighchartsSeries> prevSeries) {
    bool updateable = true;
    if (_previousSeries != null && seriesInDOM != null && 
        seriesInDOM.length == prevSeries.length) {
      Map<String, HighchartsSeries> previousIndexed = new Map<String, HighchartsSeries> ();
      _previousSeries.forEach((HighchartsSeries series) {
        if (series.id != null)
          previousIndexed[series.id] = series;  
      });  
      seriesInDOM.forEach((HighchartsSeries hcSeries) {
        updateable = updateable && previousIndexed[hcSeries.id] != null;
      });
    }
    else {
      updateable = false;
    }
    return updateable;
  }
  
  void _commitSeries (List<HighchartsSeries> seriesInDOM) {
    if (chartOptions != null) {
      if (chartOptions.series == null) {
        chartOptions.series = new List<hc.Series> ();
      }
      chartOptions.series.clear();
      seriesInDOM.forEach((HighchartsSeries highchartsSeries) {
        _seriesDictionary[highchartsSeries.id] = highchartsSeries;
        chartOptions.series.add(highchartsSeries.getSeries());
      });
    }
    _previousSeries = seriesInDOM;
  }
  
  void _updateSeries (List<HighchartsSeries> seriesInDOM) {
    seriesInDOM.forEach ((HighchartsSeries hcSeries) {
      JsObject jsSeries = jsHighchart.callMethod("get", [hcSeries.id]);
      jsSeries.callMethod("setData", [hcSeries.getSeries().getJsData()]);
    });
  }
  
  void set xAxis (hc.XAxis xAxis) {
    if (chartOptions != null) {
      chartOptions.xAxis = xAxis;
    }
    else {
      _xAxis = xAxis;
      _invalidateProperties ();
    }
  }
  
  void set yAxis (hc.YAxis yAxis) {
      if (chartOptions != null) {
        chartOptions.yAxis = yAxis;
      }
      else {
        _yAxis = yAxis;
        _invalidateProperties ();
      }
    }
  
  void chartOptionsChanged (hc.HighChart oldChartOptions) {
    _invalidateProperties ();
  }
  
  void _listenToChartOptionsChanges (hc.HighChart chartOptions) {
    chartOptions.changes.listen((records) {
      _invalidateProperties();
    });
  }
  
  void _invalidateProperties () {
    if (!_propertiesDirty) {
      _propertiesDirty = true;
      window.animationFrame.then((_) { _commitProperties(); });
    }
  }
  
  void _commitProperties () {
    if (_domReady && _propertiesDirty && chartOptions != null) {
      List<HighchartsSeries> seriesInDOM = _findSeriesInDOM ();
      if (_seriesDirty) {
        _updateableSeries = _updateable(seriesInDOM, _previousSeries);
        if (_updateableSeries) {
          _updateSeries(seriesInDOM);
        }  
        else {
          _commitSeries(seriesInDOM);
          _createChart ();
        }
        _seriesDirty = false;
      }
      
    }
    _propertiesDirty = false;
  }
  
  void _createChart () {
    if (_chartOptionsDirty)
      _commitChartOptions();
    if (_xAxis != null) {
      chartOptions.xAxis = _xAxis;
    }
    if (_yAxis != null) {
      chartOptions.yAxis = _yAxis;
    }
    _listenToChartOptionsChanges (chartOptions);
    if (chartOptions.chart == null) 
      chartOptions.chart = new hc.Chart ();
    chartOptions.chart.renderTo = mainDiv;
    mainDiv.children.clear();
    jsHighchart = new JsObject(context['Highcharts']['Chart'], [chartOptions.toJsObject()]);
    context['myChart'] = jsHighchart;
  }
  
}